"""Carrier network for SmoothRPC."""

import asyncio
import logging

from smooth_rpc.exceptions import RpcMaxSizeExceededError

_log = logging.getLogger(__name__)


class AsyncNetworkConnection:
    """
    Client network implementation.

    Converts the unix/tcp stream into messages by injecting length fields.
    """

    # arbitrary maximum message size, adjust if needed
    MAX_MSG_SIZE = 10 * 1024 * 1024  # 10MB

    # used to construct the friendly name
    connection_counter = 0

    def __init__(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter) -> None:
        """Create a network instance and connect to address."""
        self.reader = reader
        self.writer = writer

        # friendly name used in logging
        AsyncNetworkConnection.connection_counter += 1
        self.friendly_name = f"connection {AsyncNetworkConnection.connection_counter}"

    async def send_message(self, data: bytes) -> None:
        """
        Send a blob to remote.

        Insert an 8-byte big-endian length before the data blob.
        """
        if len(data) > self.MAX_MSG_SIZE:
            raise RpcMaxSizeExceededError(
                "Trying to send message larger than allowed size on %s",
                len(data),
                self.MAX_MSG_SIZE,
                self.friendly_name,
            )
        try:
            self.writer.write(len(data).to_bytes(8) + data)
            await self.writer.drain()
        except ConnectionError:
            _log.debug("Connection error on %s", self.friendly_name)
            raise

    async def recv_message(self) -> bytes:
        """
        Receive an object from remote.

        Assume an 8-byte big-endian length before a data blob.
        """
        try:
            len_bytes = await self.reader.readexactly(8)
            data_len = int.from_bytes(len_bytes)
        except asyncio.exceptions.IncompleteReadError as exc:
            if self.reader.at_eof() and len(exc.partial) == 0:
                _log.debug("Connection closed normally")
                raise

            # if we did not read a 8-byte length header our protocol is broken, re-raise exception will close connection
            _log.debug("Partial read of %d bytes on %s", len(exc.partial), self.friendly_name)
            raise

        if data_len > self.MAX_MSG_SIZE:
            raise RpcMaxSizeExceededError(
                "Trying to receive message larger than allowed size", data_len, self.MAX_MSG_SIZE
            )

        try:
            data = await self.reader.readexactly(data_len)
        except asyncio.exceptions.IncompleteReadError as exc:
            _log.debug("Did not receive enough data, expected %d, got %d", data_len, len(exc.partial))
            raise

        _log.debug("Received message of %d bytes", len(data))
        return data



"""All exceptions throw from the SmoothRPC module."""


class RpcError(Exception):
    """Base SmoothRPC exception."""


class RpcApiVersionMismatchError(RpcError):
    """Client tried to call an API endpoint that is not supported by the current api-version."""


class RpcNoSuchApiError(RpcError):
    """Host could not find the named API endpoint."""


class RpcInvalidVersionError(RpcError):
    """Host found the API endpoint, but no acceptable version(range) was found."""


class RpcProtocolError(RpcError):
    """Host or client received something unexpected."""


class RpcSerializationError(RpcError):
    """Cannot serialize(pickle) object."""


class RpcInternalError(RpcError):
    """Library failure, should not happen."""


class RpcDecoratorError(RpcError):
    """The rpc decorator was used on something that is not an async function."""


class RpcAddressError(RpcError):
    """Unrecognized address."""


class RpcMaxSizeExceededError(RpcError):
    """Message size too big."""



"""Main SmoothRPC implementations."""

import asyncio
import inspect
import logging
import pickle
import types
from collections.abc import Callable, Generator
from dataclasses import dataclass
from functools import partial, wraps
from typing import NamedTuple, ParamSpec, TypeVar, cast

from smooth_rpc.exceptions import (
    RpcApiVersionMismatchError,
    RpcDecoratorError,
    RpcInternalError,
    RpcInvalidVersionError,
    RpcNoSuchApiError,
    RpcProtocolError,
    RpcSerializationError,
)
from smooth_rpc.network import AsyncNetworkConnection

_log = logging.getLogger(__name__)


class VersionRange(NamedTuple):
    """A closed range."""

    min_version: int | None
    max_version: int | None

    def __contains__(self, version: object) -> bool:
        """Return if the given version is in [min_version, max_version]."""
        if not isinstance(version, int):
            raise RpcInternalError("Can only compare int's")
        return (self.min_version is None or self.min_version <= version) and (
            self.max_version is None or version <= self.max_version
        )


@dataclass
class ApiFunctionSpec:
    """Function attributes, as stored by 'rpc' decorator."""

    name: str | None
    version_range: VersionRange


@dataclass
class ApiCall:
    """Network object sent from client to host."""

    func_name: str
    api_version: int
    args: tuple
    kwargs: dict

    def __str__(self) -> str:
        """Represent this call object as string. Used by tests."""
        return f"{self.func_name}(args={self.args!r}, kwargs={self.kwargs!r}, api_version={self.api_version})"


Param = ParamSpec("Param")
RetType = TypeVar("RetType")


def _create_call_wrapper(
    connection: AsyncNetworkConnection,
    func_name: str,
    api_version: int,
    method_self: object,
    func: Callable[Param, RetType],
) -> Callable[Param, RetType]:
    @wraps(func)
    async def do_call(_self: object, *args: tuple, **kwargs: dict) -> RetType:
        # Construct call object, pickle and send
        call_obj = ApiCall(func_name, api_version, args, kwargs)
        _log.info("Sending RPC call %s", call_obj)
        in_data = pickle.dumps(call_obj)
        await connection.send_message(in_data)

        # Receive response and unpickle
        out_data = await connection.recv_message()
        try:
            out = pickle.loads(out_data)  # noqa: S301, use pickle anyway, for lack of a safe alternative
        except (pickle.UnpicklingError, IndexError, AttributeError) as exc:
            # server should never send us something that we cannot unpickle
            raise RpcProtocolError("Received response that could not be unpickled") from exc
        _log.info("Call %s returned %r", func_name, out)

        if isinstance(out, Exception):
            raise out  # re-raise received exception
        return cast(RetType, out)

    return types.MethodType(do_call, method_self)


def _create_exception_wrapper(method_self: object, func: Callable[Param, RetType]) -> Callable[Param, RetType]:
    @wraps(func)
    async def do_throw(_self: object, *_args: tuple, **_kwargs: dict) -> RetType:
        raise RpcApiVersionMismatchError("Client API disabled due to api version restriction.")

    return types.MethodType(do_throw, method_self)


def _iterate_functions(command_object: object) -> Generator[tuple[str, str, VersionRange, Callable]]:
    for superclass in command_object.__class__.__mro__:
        for func_object_name, func in superclass.__dict__.items():
            api_spec = getattr(func, "__is_part_of_smooth_api", None)
            if api_spec is None:
                continue
            if not inspect.iscoroutinefunction(func):
                raise RpcDecoratorError("Not an async function")
            if not isinstance(api_spec, ApiFunctionSpec):
                raise RpcDecoratorError("Decorator malfunction")

            func_name = api_spec.name or func_object_name
            yield (func_object_name, func_name, api_spec.version_range, func)


def instrument_client_commands(connection: AsyncNetworkConnection, api_version: int, *command_objects: object) -> None:
    """Replace all functions in command_object that are marked with the api decorator with remote calls."""
    for command_object in command_objects:
        for org_func_name, func_name, version_range, func in _iterate_functions(command_object):
            if api_version in version_range:
                wrapper = _create_call_wrapper(connection, func_name, api_version, command_object, func)
            else:
                wrapper = _create_exception_wrapper(command_object, func)
            setattr(command_object, org_func_name, wrapper)


class SmoothRPCHost:
    """Host for the RPC API."""

    def __init__(self) -> None:
        """
        Create a RPC Host.

        Actually, just a dict of all registered commands and a connection-acceptor function.
        """
        self.commands: dict[str, list[tuple[VersionRange, Callable]]] = {}

    def register_commands(self, command_object: object) -> None:
        """Add all rpc-decorator marked functions in command_object to host command list."""
        count = 0
        for _org_func_name, func_name, version_range, func in _iterate_functions(command_object):
            func_bound = partial(func, command_object)
            self.commands.setdefault(func_name, []).append((version_range, func_bound))
            count += 1
        if count == 0:
            raise RpcDecoratorError("No decorated functions found")
        _log.info("Registered %d API endpoints", count)

    async def _call_command(self, call_obj: ApiCall) -> object:
        _log.info("Received RPC call %s", call_obj)

        try:
            version_list = self.commands[call_obj.func_name]
        except KeyError as err:
            raise RpcNoSuchApiError("No such api", call_obj.func_name) from err

        # find the first api that supports the requested api version
        matching_func: Callable
        for version_range, func in version_list:
            if call_obj.api_version in version_range:
                matching_func = func
                break
        else:
            raise RpcInvalidVersionError("Api endpoint does not support api version", call_obj.api_version)

        out = await matching_func(*call_obj.args, **call_obj.kwargs)
        _log.info("Call %s returned %r", call_obj.func_name, out)
        return out

    def _pack_result(self, obj: object) -> bytes:
        """Return the object pickled, or a pickled exception that it cannot be pickled."""
        try:
            return pickle.dumps(obj)
        except (pickle.PicklingError, AttributeError) as exc:
            return pickle.dumps(RpcSerializationError("Could not pickle object", exc))

    def _unpack_call(self, data: bytes) -> ApiCall:
        """Unpickle a received object and check the type."""
        try:
            call_obj = pickle.loads(data)  # noqa: S301, use pickle anyway, for lack of a safe alternative
        except (pickle.UnpicklingError, IndexError, AttributeError) as exc:
            raise RpcProtocolError("Failed to unpickle received data") from exc
        if not isinstance(call_obj, ApiCall):
            raise RpcProtocolError("Invalid object received")
        return call_obj

    async def _handle_one_command(self, connection: AsyncNetworkConnection) -> None:
        """Handle exactly one RPC call, receive/call/send."""
        in_data = await connection.recv_message()  # network exceptions propagate upwards
        call_obj = self._unpack_call(in_data)  # protocol errors propagate upwards

        try:
            out = await self._call_command(call_obj)
        except Exception as exc:  # noqa: BLE001, blind-catch exception
            _log.warning("Caught exception in call, returning to client: %r", exc)
            out = exc

        data = self._pack_result(out)
        await connection.send_message(data)  # network exceptions propagate upwards

    async def accept_connection(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter) -> None:
        """Use an existing connection for answering requests. Formatted for use with asyncio.Server."""
        connection = AsyncNetworkConnection(reader, writer)
        _log.info("Hosting client %s", connection.friendly_name)
        try:
            while True:
                await self._handle_one_command(connection)
        except asyncio.exceptions.IncompleteReadError as exc:
            # suppress exception on normal close
            if not reader.at_eof() or len(exc.partial) > 0:
                raise
        finally:
            _log.info("Hosting client %s closed", connection.friendly_name)


def api(*, func_name: str | None = None, min_version: int | None = None, max_version: int | None = None) -> Callable:
    """
    Mark a function as part of the API, use as decorator.

    Arguments are:
    - func_name, override the api name (defaults to function name).
    - min_version, first api version that supports this call.
    - max_version, last api version that supports this call.
    """

    def api_decorator(func: Callable[Param, RetType]) -> Callable[Param, RetType]:
        # Set a marker on the function so it can be instrumented later.
        # Neither mypy nor flake8 like us accessing a private non-existing  member
        spec = ApiFunctionSpec(func_name, VersionRange(min_version, max_version))
        func.__is_part_of_smooth_api = spec  # type: ignore[attr-defined]  # noqa: SLF001
        return func

    return api_decorator



"""SmoothRPC Module."""

import asyncio
from pathlib import Path

from smooth_rpc.exceptions import RpcAddressError, RpcError
from smooth_rpc.network import AsyncNetworkConnection
from smooth_rpc.smoothrpc import SmoothRPCHost, api, instrument_client_commands

__all__ = [
    "api",
    "AsyncNetworkConnection",
    "host_forever",
    "init_remote_rpc",
    "RpcError",
    "SmoothRPCHost",
]


async def host_forever(address: str, *commands: object) -> None:  # pragma: no cover
    """
    Listen on 'address and host all rpc-marked functions from commands.

    Note that this is a convenience function, that does not support stopping the host.
    """
    rpc = SmoothRPCHost()
    for command_object in commands:
        rpc.register_commands(command_object)

    match address:
        case address if address.startswith("ipc://"):
            path = Path(address[6:])
            host = await asyncio.start_unix_server(rpc.accept_connection, path=path)
        case address if address.startswith("tcp://"):
            # FIXME: test the tcp client
            ip_host, tcp_port = address[6:].split(":", 2)
            host = await asyncio.start_server(rpc.accept_connection, host=ip_host, port=int(tcp_port))
        case _:
            raise RpcAddressError("Unrecognized address", address)

    await host.serve_forever()


async def init_remote_rpc(
    address: str, *commands: object, api_version: int = 1
) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:  # pragma: no cover
    """
    Replace all rpc-marked functions from commands with remote calls to address.

    A connection to 'address' is opened. The connection object is returned, so it
    can be used to explicitly close the connection.
    """
    match address:
        case address if address.startswith("ipc://"):
            path = Path(address[6:])
            reader, writer = await asyncio.open_unix_connection(path=path)
        case address if address.startswith("tcp://"):
            # FIXME: test the tcp client
            ip_host, tcp_port_str = address[6:].split(":", 2)
            tcp_port = int(tcp_port_str)
            reader, writer = await asyncio.open_connection(host=ip_host, port=tcp_port)
        case _:
            raise RpcAddressError("Unrecognized address", address)
    connection = AsyncNetworkConnection(reader, writer)
    instrument_client_commands(connection, api_version, *commands)
    return reader, writer
